#ifndef EXPERIMENTSCOPE_H
#define EXPERIMENTSCOPE_H

#include "util/ksatinstance.h"
#include <memory>
#include <functional>

namespace satlab
{

class ExperimentScope
{
protected:
    
    class InstanceIncrementer
    {
    public:
        friend bool operator==(const ExperimentScope::InstanceIncrementer& lhs,
                               const ExperimentScope::InstanceIncrementer& rhs);
        
        InstanceIncrementer() = default;
        virtual ~InstanceIncrementer() = default;
        
        InstanceIncrementer(const InstanceIncrementer& other) = default;
        InstanceIncrementer(InstanceIncrementer&& other) = default;
        
        InstanceIncrementer& operator=(const InstanceIncrementer& other) = default;
        InstanceIncrementer& operator=(InstanceIncrementer&& other) = default;
        
        virtual void next();
        
        virtual KSATinstance* operator->();
        
        virtual KSATinstance& operator*();
        
    private:
        int step;
        
    protected:
        KSATinstance instance;
    };
    
    using P_INCR_T = std::shared_ptr<InstanceIncrementer>;
    
public:
    class iterator 
    {
    public:
        friend class ExperimentScope;
        
        using iterator_category = std::forward_iterator_tag;
        using value_type = KSATinstance;
        using difference_type = int;
        using pointer = KSATinstance*;
        using reference = KSATinstance&;
        
        iterator() = default;
        virtual ~iterator() = default;
        
        iterator(const iterator& other) = default;
        iterator& operator=(const iterator& other) = default;
        
        virtual bool operator==(const iterator&other);
        virtual bool operator!=(const iterator&other);
        
        virtual reference operator*();
        virtual pointer operator->();
        
        virtual iterator& operator++();
        virtual iterator operator++(int);
        
    private:
        using CMP_FNC_T = std::function<bool(const P_INCR_T& lhs,
                                             const P_INCR_T& rhs)>;
                                             
        iterator(P_INCR_T&& pIncr, CMP_FNC_T&& cmp);
        
        CMP_FNC_T cmp;
    
        P_INCR_T pIncr;
    };
    
    virtual iterator begin() noexcept = 0;
    
    virtual iterator end() noexcept = 0;
    
protected:
    iterator createIterator(P_INCR_T&& pIncr, iterator::CMP_FNC_T&& cmp);
    
};

}

#endif // EXPERIMENTSCOPE_H
