#ifndef MONGODBEXPERIMENTSCOPE_H
#define MONGODBEXPERIMENTSCOPE_H

#include "util/experimentscope.h"

#include "mongocxx/cursor.hpp"
#include "mongocxx/database.hpp"

#include <string>

namespace satlab
{

class MongoDBExperimentScope : public ExperimentScope
{
public:
    using BASE = ExperimentScope;
    
    class InstanceIncrementer : public ExperimentScope::InstanceIncrementer
    {
    public:
        friend class MongoDBExperimentScope;
        
        InstanceIncrementer() = default;
        virtual ~InstanceIncrementer() = default;
        
        InstanceIncrementer(const InstanceIncrementer& other) = default;
        InstanceIncrementer(InstanceIncrementer&& other) = default;
        
        InstanceIncrementer& operator=(const InstanceIncrementer& other) = default;
        InstanceIncrementer& operator=(InstanceIncrementer&& other) = default;
        
        InstanceIncrementer(mongocxx::cursor::iterator&& cursorIt);
        
        virtual void next() override;
        
        virtual KSATinstance* operator->() override;
        
        virtual KSATinstance& operator*() override;
        
    private:
        void updateInstance();
        
        bool instanceIsUpdated = false;
        mongocxx::cursor::iterator cursorIt;
    };
    
public:
    MongoDBExperimentScope() = default;
    virtual ~MongoDBExperimentScope() = default;
    
    MongoDBExperimentScope(const MongoDBExperimentScope& other) = default;
    MongoDBExperimentScope(MongoDBExperimentScope&& other) = default;
    
    MongoDBExperimentScope& operator=(const MongoDBExperimentScope& other) = default;
    MongoDBExperimentScope& operator=(MongoDBExperimentScope&& other) = default;
    
    MongoDBExperimentScope(mongocxx::cursor&& cursor);
    MongoDBExperimentScope(const mongocxx::database& db, const std::string& scopeName);
    
    virtual BASE::iterator begin() noexcept override;
    virtual BASE::iterator end() noexcept override;
    
private:
    using P_INCR_T = BASE::P_INCR_T;
    
    mongocxx::cursor getCursor(const mongocxx::database& db,
                               const std::string& scopeName);
    
    mongocxx::pipeline createScopePipeline(const std::string& scopeName);
    
    mongocxx::cursor cursor;
};

}

#endif // MONGODBEXPERIMENTSCOPE_H
